extern crate kiss3d;
extern crate nalgebra as na;

use cdl2d::transformation::vhacd::{VHACDParameters as VHACDParameters2, VHACD as VHACD2};
use cdl2d::transformation::voxelization::FillMode as FillMode2;
use cdl3d::transformation::vhacd::{VHACDParameters, VHACD};
use kiss3d::camera::ArcBall;
use kiss3d::event::{Action, WindowEvent};
use kiss3d::light::Light;
use kiss3d::loader::obj;
use kiss3d::planar_camera::{FixedView, PlanarCamera};
use kiss3d::scene::PlanarSceneNode;
use kiss3d::window::Window;
use na::{Point2, Point3, Translation2, Vector2, Vector3};
use ncollide3d::procedural::{IndexBuffer, TriMesh};
use ncollide3d::shape::Triangle;
use std::path::Path;

fn main() {
    const ENABLE_2D_DEMO: bool = false;
    let mut window = Window::new("Kiss3d: obj");
    window.set_light(Light::StickToCamera);

    // Teapot
    let obj_path = Path::new("media/teapot/teapot.obj");
    let mtl_path = Path::new("media/teapot");
    let teapot = obj::parse_file(obj_path, mtl_path, "teapot").unwrap();
    dbg!(teapot.len());

    let mut all_points = Vec::new();
    let mut all_triangles = Vec::new();

    for (_, mesh, _) in teapot.into_iter() {
        let base_i = all_points.len() as u32;
        let mut trimesh = mesh.to_trimesh().unwrap();
        let mut triangles: Vec<_> = trimesh
            .flat_indices()
            .chunks(3)
            .map(|i| [i[0] + base_i, i[1] + base_i, i[2] + base_i])
            .collect();
        all_points.append(&mut trimesh.coords);
        all_triangles.append(&mut triangles);
    }

    /*
    let voxels = VoxelSet::voxelize(&Isometry3::identity(), &all_points, &all_triangles, 70, FillMode::FloodFill);

    for voxel in voxels.voxels() {
        let pos = voxels.origin + voxel.coords.coords.map(|k| k as f32) * voxels.scale;
        window.add_cube(voxels.scale, voxels.scale, voxels.scale)
            .set_local_translation(Translation3::new(pos.x, pos.y, pos.z));
    }
     */

    let start = std::time::Instant::now();
    let mut params = VHACDParameters::default();
    let vhacd = VHACD::decompose(&params, &all_points, &all_triangles, true);

    /*
    for part in vhacd.voxel_parts() {
        for voxel in part.voxels() {
            let pos = part.origin + voxel.coords.coords.map(|k| k as f32) * part.scale;
            window.add_cube(part.scale, part.scale, part.scale)
                .set_local_translation(Translation3::new(pos.x, pos.y, pos.z));
        }
    }
     */

    let end = std::time::Instant::now();
    println!(
        "The convex decomposition is done: {} parts in {} seconds.",
        vhacd.voxel_parts().len(),
        (end - start).as_secs_f32()
    );

    let convex_hulls = vhacd.compute_exact_convex_hulls(&all_points, &all_triangles);
    // let convex_hulls = vhacd.compute_convex_hulls(1);

    for (vertices, indices) in convex_hulls {
        let color = Point3::new(rand::random(), rand::random(), rand::random());
        let mut flat_vertices = vec![];

        for idx in &indices {
            flat_vertices.push(vertices[idx[0] as usize]);
            flat_vertices.push(vertices[idx[1] as usize]);
            flat_vertices.push(vertices[idx[2] as usize]);
        }

        for flat_vertices in flat_vertices.chunks((u16::MAX as usize / 3) * 3) {
            let mesh = TriMesh::new(flat_vertices.to_vec(), None, None, None);
            let mut node = window.add_trimesh(mesh, Vector3::repeat(1.0));
            node.set_color(color.x, color.y, color.z);
            // node.set_surface_rendering_activation(false);
            // node.set_lines_width(10.0);
        }
    }

    // for vertices in vhacd.compute_primitive_intersections(&all_points, &all_triangles) {
    //     let color = Point3::new(rand::random(), rand::random(), rand::random());
    //     for vertices in vertices.chunks((u16::MAX as usize / 3) * 3) {
    //         let mesh = TriMesh::new(vertices.to_vec(), None, None, None);
    //         window
    //             .add_trimesh(mesh, Vector3::repeat(1.0))
    //             .set_color(color.x, color.y, color.z);
    //     }
    // }

    /*
     * 2D demo.
     */
    if ENABLE_2D_DEMO {
        let circle1 = cdl2d::shape::Ball::new(10.0);
        let circle2 = cdl2d::shape::Ball::new(4.0);
        let mut points1 = circle1.to_polyline(20);
        let mut points2 = circle2.to_polyline(20);

        let mut segments1: Vec<_> = (0u32..points1.len() as u32)
            .map(|i| [i, (i + 1) % points1.len() as u32])
            .collect();

        let mut segments2: Vec<_> = (0u32..points2.len() as u32)
            .map(|i| {
                [
                    i + points1.len() as u32,
                    ((i + 1) % points2.len() as u32) + points1.len() as u32,
                ]
            })
            .collect();
        points1.append(&mut points2);
        segments1.append(&mut segments2);

        for i in 1..3 {
            let circle3 = cdl2d::shape::Ball::new(10.0);
            let mut points3 = circle3.to_polyline(20);
            points3.iter_mut().for_each(|p| p.y += i as f32 * 3.0);

            let mut segments3: Vec<_> = (0u32..points3.len() as u32)
                .map(|i| {
                    [
                        i + points1.len() as u32,
                        ((i + 1) % points3.len() as u32) + points1.len() as u32,
                    ]
                })
                .collect();
            points1.append(&mut points3);
            segments1.append(&mut segments3);
        }

        let mut params = VHACDParameters2::default();
        params.fill_mode = FillMode2::FloodFill {
            detect_cavities: true,
            detect_self_intersections: true,
        };
        params.concavity = 0.01;
        // params.fill_mode = FillMode2::SurfaceOnly;
        let vhacd = VHACD2::decompose(&params, &points1, &segments1, true);
        println!("Parts 2D: {}", vhacd.voxel_parts().len());

        // for part in vhacd.voxel_parts() {
        //     let color = (rand::random(), rand::random(), rand::random());
        //     for voxel in part.voxels() {
        //         let pos = (part.origin + voxel.coords.coords.map(|k| k as f32) * part.scale) * 10.0;
        //         let mut node = window.add_rectangle(part.scale * 10.0, part.scale * 10.0);
        //         node.set_local_translation(Translation2::new(pos.x, pos.y));
        //         node.set_color(color.0, color.1, color.2);
        //     }
        // }

        for points in vhacd.compute_exact_convex_hulls(&points1, &segments1) {
            if points.len() > 1 {
                window
                    .add_convex_polygon(points, Vector2::repeat(10.0))
                    .set_color(rand::random(), rand::random(), rand::random());
            }
        }

        let mut cursor_pos = (0.0, 0.0);
        let mut points = vec![];
        let mut indices = vec![];
        let mut nodes: Vec<PlanarSceneNode> = vec![];
        let mut camera = ArcBall::new(Point3::new(0.0, 0.0, -100.0), Point3::origin());
        let mut planar_camera = FixedView::new();

        while window.render_with_cameras(&mut camera, &mut planar_camera) {
            for event in window.events().iter() {
                match event.value {
                    WindowEvent::CursorPos(x, y, _) => {
                        cursor_pos = (x, y);
                    }
                    WindowEvent::MouseButton(_, Action::Release, _) => {
                        if !points.is_empty() {
                            if !indices.is_empty() {
                                indices.pop();
                            }
                            indices.push([points.len() as u32 - 1, points.len() as u32]);
                            indices.push([points.len() as u32, 0]);
                        }

                        let sz = window.size();
                        let pt = planar_camera.unproject(
                            &Point2::new(cursor_pos.0 as f32, cursor_pos.1 as f32),
                            &Vector2::new(sz.x as f32, sz.y as f32),
                        );
                        points.push(pt);

                        for mut node in nodes.drain(..) {
                            window.remove_planar_node(&mut node);
                        }

                        if points.len() >= 3 {
                            let mut params = VHACDParameters2::default();
                            params.fill_mode = FillMode2::FloodFill {
                                detect_cavities: true,
                                detect_self_intersections: true,
                            };
                            // params.fill_mode = FillMode2::SurfaceOnly;
                            let vhacd = VHACD2::decompose(&params, &points, &indices, true);
                            println!("Parts 2D: {}", vhacd.voxel_parts().len());

                            // for part in vhacd.voxel_parts() {
                            //     let color = (rand::random(), rand::random(), rand::random());
                            //     for voxel in part.voxels() {
                            //         let pos = part.origin
                            //             + voxel.coords.coords.map(|k| k as f32) * part.scale;
                            //         let mut node = window.add_rectangle(part.scale, part.scale);
                            //         node.set_local_translation(Translation2::new(pos.x, pos.y));
                            //         node.set_color(color.0, color.1, color.2);
                            //         nodes.push(node);
                            //     }
                            // }

                            for points in vhacd.compute_exact_convex_hulls(&points, &indices) {
                                if points.len() > 1 {
                                    let mut node =
                                        window.add_convex_polygon(points, Vector2::repeat(1.0));
                                    node.set_color(rand::random(), rand::random(), rand::random());
                                    nodes.push(node);
                                }
                            }
                        }
                    }
                    _ => {}
                }
            }
        }
    } else {
        while window.render() {}
    }
}
